from __future__ import annotations

import re
from collections import defaultdict
from ctypes import c_float
from sys import stderr
from typing import Any, Callable, Dict, Iterable, Optional, Set, Tuple, Union

from .errors import MoreThanOneAltAllele, NotExactlyOneValue


def float32(val: str) -> float:
    return c_float(float(val)).value


# If NoValue inherits from str, re.search("something", NoValue()) does not error
# but just comes up empty-handed, which is convenient behaviour.
# This way, we do not have to special case / monkey patch / wrap the regex module.
class NoValue(str):
    warnings: Set[str] = set()

    def __lt__(self, other):
        return False

    def __gt__(self, other):
        return False

    def __le__(self, other):
        return False

    def __ge__(self, other):
        return False

    def __eq__(self, other):
        return False

    def __ne__(self, other):
        return True

    def __bool__(self):
        return False

    def __str__(self):
        # nonexistent fields will result in an empty string
        # should be configurable in upcoming versions
        return ""

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def __hash__(self) -> int:
        return super().__hash__()

    def __getattr__(self, item):
        if item not in self.warnings:
            self.warnings.add(item)
            print(
                f"Warning: Trying to access non-existent attribute '{item}' of NoValue."
                " Returning NA instead.\n"
                "This warning will only be printed once per attribute.\n"
                "It either indicates a typo in the attribute name or "
                "the access of an attribute of a field with a custom type "
                "which is empty/has no value.",
                file=stderr,
            )
        return self


NA = NoValue()


class InfoTuple:
    """A container that lazily evaluates None to NA in case of access."""

    def __init__(self, values):
        self.values = values

    def __getitem__(self, spec):
        values = self.values.__getitem__(spec)
        if isinstance(values, tuple) and any(v is None for v in values):
            return tuple((NA if v is None else v) for v in values)
        if values is None:
            return NA
        return values

    def __str__(self):
        return self.values.__str__()

    def __repr__(self):
        return f"{self.__class__.__name__}(values={self.values!r})"

    def __len__(self):
        return len(self.values)

    def __eq__(self, other):
        value_type = type(self.values)
        if isinstance(other, InfoTuple) and isinstance(other.values, value_type):
            return self.values == other.values
        elif isinstance(other, value_type):
            return self.values == other
        else:
            raise TypeError(
                f"Incomparable: "
                f"{type(self)} (value type: {type(self.values)}), "
                f"{type(other)}"
                + (
                    f" (value type: {type(other.values)})"
                    if isinstance(other, InfoTuple)
                    else ""
                )
            )

    def __hash__(self):
        return hash(self.values)


IntFloatStr = Union[int, float, str]
NvIntFloatStr = Union[IntFloatStr, NoValue]
NvInt = Union[int, NoValue]
NvFloat = Union[float, NoValue]


def type_info(
    value, number=".", field=None, record_idx=None
) -> Union[NvIntFloatStr, InfoTuple]:
    if value is None:
        return NA
    if number == "0":
        return NA if value is None else value
    if number == "A":
        if len(value) != 1:
            raise MoreThanOneAltAllele()
        return value[0] if value[0] is not None else NA
    if number == "R":
        if len(value) != 2:
            raise MoreThanOneAltAllele()
        return InfoTuple(value)
    if number == "1":
        if isinstance(value, tuple):
            if len(value) != 1:
                raise NotExactlyOneValue(field, len(value), record_idx)
            return value[0] if value[0] is not None else NA
        return value if value is not None else NA
    if isinstance(value, tuple):
        return InfoTuple(value)
    else:
        return InfoTuple((value,))


class PosRange:
    def __init__(self, start: NvInt, end: NvInt, raw: str):
        self.start = start
        self.end = end
        self._raw = raw

    raw = property(lambda self: self._raw, None, None)

    @classmethod
    def from_snpeff_str(cls, value: str) -> PosRange:
        pos, length = [int(v.strip()) for v in value.split("/")]
        return cls(pos, pos + length, value)

    @classmethod
    def from_vep_str(cls, value: str) -> PosRange:
        if "-" in value:
            s, e = map(str.strip, value.split("-", 1))
            start: NvInt = NA if s == "?" else int(s)
            end: NvInt = NA if e == "?" else int(e)
            return cls(start, end, value)
        elif len(value) > 0:
            start = int(value.strip())
            return cls(start, start + 1, value)
        else:
            return cls(NA, NA, value)

    @property
    def length(self):
        if self.end is not NA and self.start is not NA:
            return self.end - self.start
        else:
            return NA

    def __str__(self):
        return f"(start: {self.start}, end: {self.end}, length: {self.length})"

    def __repr__(self):
        return f"{self.__class__.__name__}(start={self.start!r}, end={self.end!r})"


def na_func() -> NoValue:
    return NA


AnnotationType = Union[IntFloatStr, Iterable[IntFloatStr], NoValue]


class AnnotationEntry:
    def __init__(
        self,
        name: str,
        typefunc: Callable[[str], Any] = str,
        nafunc: Callable[[], Any] = na_func,
        description: Optional[str] = None,
    ):
        self._name = name
        self._typefunc = typefunc
        self._nafunc = nafunc
        self._description = description

    @property
    def name(self):
        return self._name

    def convert(self, value: str) -> Any:
        if value:
            return self._typefunc(value)
        else:
            return self._nafunc()

    def description(self):
        return self._description

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"name={self._name!r}, "
            f"typefunc={self._typefunc!r}, "
            f"nafunc={self._nafunc!r}, "
            f"description={self._description!r}"
            f")"
        )


class AnnotationListEntry(AnnotationEntry):
    def __init__(
        self,
        name: str,
        sep: str,
        typefunc: Optional[Callable[[str], Any]] = None,
        inner_typefunc: Callable[[str], Any] = lambda x: x,
        **kwargs,
    ):
        typefunc = (
            typefunc
            if typefunc
            else lambda v: [inner_typefunc(x.strip()) for x in v.split(sep)]
        )
        super().__init__(name, typefunc, nafunc=lambda: [], **kwargs)


class AnnotationListDictEntry(AnnotationEntry):
    def __init__(self, name: str, nafunc=lambda: None, **kwargs):
        def typefunc(x):
            key_value_tuples = map(lambda v: v.split(":", maxsplit=1), x.split("&"))
            mapping = defaultdict(list)
            for key, value in key_value_tuples:
                mapping[key].append(value)
            # make sure non-existent keys do not return list() on __getitem__ calls
            # in the filter expression (without actually building a new dict from
            # the defaultdict)
            mapping.default_factory = nafunc
            return mapping

        super().__init__(name=name, typefunc=typefunc, **kwargs)


class RangeTotal(object):
    def __init__(self, r: range, total: int, raw: str):
        self.range = r
        self.total = total
        self._raw = raw

    raw = property(lambda self: self._raw, None, None)

    @classmethod
    def from_vep_string(cls, value: str) -> RangeTotal:
        v = value.split("/")
        r = [int(s) for s in v[0].split("-")]
        if len(r) == 1:
            return cls(range(r[0], r[0] + 1), int(v[1]), value)
        elif len(r) == 2:
            return cls(range(r[0], r[1] + 1), int(v[1]), value)
        else:
            raise ValueError(
                "Found more than two values separated by '-', "
                "expected only a single int, or two ints separated by '-'."
            )

    def __str__(self):
        if len(self.range) == 1:
            return f"number / total: {self.range.start} / {self.total}"
        else:
            return (
                f"range / total: "
                f"{self.range.start} - {self.range.stop - 1} / {self.total}"
            )

    def __repr__(self):
        return (
            f"{self.__class__.__name__}"
            f"(range=range({self.range.start!r}, {self.range.stop!r}), "
            f"total={self.total!r})"
        )


class AnnotationRangeTotalEntry(AnnotationEntry):
    def __init__(self, name: str, **kwargs):
        super().__init__(name, typefunc=RangeTotal.from_vep_string, **kwargs)


class NumberTotal(object):
    def __init__(self, number: int, total: int, raw: str):
        self.number = number
        self.total = total
        self._raw = raw

    raw = property(lambda self: self._raw, None, None)

    @classmethod
    def from_vep_string(cls, value: str) -> NumberTotal:
        v = value.split("/")
        return cls(int(v[0]), int(v[1]), value)

    def __str__(self):
        return f"number / total: {self.number} / {self.total}"

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(number={self.number!r}, total={self.total!r})"
        )


class AnnotationNumberTotalEntry(AnnotationEntry):
    def __init__(self, name: str, **kwargs):
        super().__init__(name, typefunc=NumberTotal.from_vep_string, **kwargs)


PREDICTION_SCORE_REGEXP: re.Pattern = re.compile(r"(.*)\((.*)\)")


class AnnotationPredictionScoreEntry(AnnotationEntry):
    def __init__(self, name: str, **kwargs):
        def typefunc(x: str) -> Dict[str, float]:
            match = PREDICTION_SCORE_REGEXP.findall(x)[0]
            return {match[0]: float32(match[1])}

        super().__init__(name, typefunc=typefunc, nafunc=lambda: dict(), **kwargs)


class DefaultAnnotationEntry(AnnotationEntry):
    def __init__(self, name: str):
        super().__init__(name)


class AnnotationTyper:
    def __init__(self, mapping: Dict[str, AnnotationEntry]):
        self._mapping = mapping

    def get_entry(self, key: str) -> AnnotationEntry:
        entry = self._mapping.get(key)
        if not entry:
            print(
                f"No type information available for '{key}', defaulting to `str`. "
                f"If you would like to have a custom type for this, "
                f"please consider filing an issue at "
                f"https://github.com/vembrane/vembrane/issues",
                file=stderr,
            )
            self._mapping[key] = DefaultAnnotationEntry(key)
            entry = self._mapping[key]
        return entry

    def convert(self, key: str, value: str) -> Tuple[str, AnnotationType]:
        entry = self.get_entry(key)
        return entry.name, entry.convert(value)


KNOWN_ANN_TYPE_MAP_SNPEFF = {
    "Allele": AnnotationEntry("Allele"),
    "Annotation": AnnotationEntry("Annotation"),
    "Annotation_Impact": AnnotationEntry("Annotation_Impact"),
    "Gene_Name": AnnotationEntry("Gene_Name"),
    "Gene_ID": AnnotationEntry("Gene_ID"),
    "Feature_Type": AnnotationEntry("Feature_Type"),
    "Feature_ID": AnnotationEntry("Feature_ID"),
    "Transcript_BioType": AnnotationEntry("Transcript_BioType"),
    "Rank": AnnotationEntry("Rank"),
    "HGVS.c": AnnotationEntry("HGVS.c"),
    "HGVS.p": AnnotationEntry("HGVS.p"),
    "cDNA.pos / cDNA.length": AnnotationEntry(
        "cDNA.pos / cDNA.length", PosRange.from_snpeff_str
    ),
    "CDS.pos / CDS.length": AnnotationEntry(
        "CDS.pos / CDS.length", PosRange.from_snpeff_str
    ),
    "AA.pos / AA.length": AnnotationEntry(
        "AA.pos / AA.length", PosRange.from_snpeff_str
    ),
    "Distance": AnnotationEntry("Distance", str),
    "ERRORS / WARNINGS / INFO": AnnotationListEntry(
        "ERRORS / WARNINGS / INFO",
        "/",
    ),
}

# fields, types and description taken from:
# https://www.ensembl.org/info/docs/tools/vep/vep_formats.html#other_fields
KNOWN_ANN_TYPE_MAP_VEP = {
    "Location": AnnotationEntry(
        "Location",
        description="In standard coordinate format (chr:start or chr:start-end)",
    ),
    "Allele": AnnotationEntry(
        "Allele", description="The variant allele used to calculate the consequence"
    ),
    "Gene": AnnotationEntry("Gene", description="Ensembl stable ID of affected gene"),
    "Feature": AnnotationEntry("Feature", description="Ensembl stable ID of feature"),
    "Feature_type": AnnotationEntry(
        "Feature_type",
        description="Type of feature. "
        "Currently one of Transcript, RegulatoryFeature, MotifFeature.",
    ),
    "Consequence": AnnotationEntry(
        "Consequence", description="Consequence type of this variant"
    ),
    "cDNA_position": AnnotationEntry("cDNA_position", PosRange.from_vep_str),
    "CDS_position": AnnotationEntry("CDS_position", PosRange.from_vep_str),
    "Protein_position": AnnotationEntry("Protein_position", PosRange.from_vep_str),
    "HGSVc": AnnotationEntry("HGSVc"),
    "HGSVp": AnnotationEntry("HGSVp"),
    "REF_ALLELE": AnnotationEntry("REF_ALLELE", description="The reference allele"),
    "IMPACT": AnnotationEntry(
        "IMPACT", description="The impact modifier for the consequence type"
    ),
    "SYMBOL": AnnotationEntry("SYMBOL", description="The gene symbol"),
    "VARIANT_CLASS": AnnotationEntry(
        "VARIANT_CLASS", description="Sequence Ontology variant class"
    ),
    "SYMBOL_SOURCE": AnnotationEntry(
        "SYMBOL_SOURCE", description="The source of the gene symbol"
    ),
    "STRAND": AnnotationEntry(
        "STRAND",
        int,
        description="The DNA strand (1 or -1) on which the transcript/feature lies",
    ),
    "ENSP": AnnotationEntry(
        "ENSP", description="The Ensembl protein identifier of the affected transcript"
    ),
    "FLAGS": AnnotationListEntry(
        "FLAGS",
        sep="&",
        description="Transcript quality flags: "
        "cds_start_NF: CDS 5' incomplete, cds_end_NF: CDS 3' incomplete",
    ),
    "SWISSPROT": AnnotationEntry(
        "SWISSPROT",
        description="Best match UniProtKB/Swiss-Prot accession of protein product",
    ),
    "TREMBL": AnnotationEntry(
        "TREMBL", description="Best match UniProtKB/TrEMBL accession of protein product"
    ),
    "UNIPARC": AnnotationEntry(
        "UNIPARC", description="Best match UniParc accession of protein product"
    ),
    "HGVSc": AnnotationEntry("HGVSc", description="The HGVS coding sequence name"),
    "HGVSp": AnnotationEntry("HGVSp", description="The HGVS protein sequence name"),
    "HGVSg": AnnotationEntry("HGVSg", description="The HGVS genomic sequence name"),
    "HGVS_OFFSET": AnnotationEntry(
        "HGVS_OFFSET",
        int,
        description="Indicates by how many bases the HGVS notations for this variant "
        "have been shifted",
    ),
    # "NEAREST": AnnotationEntry("NEAREST",
    #               description="Identifier(s) of nearest transcription start site"),
    "SIFT": AnnotationPredictionScoreEntry(
        "SIFT",
        description="The SIFT prediction and/or score,"
        " with both given as prediction(score)",
    ),
    "PolyPhen": AnnotationPredictionScoreEntry(
        "PolyPhen", description="The PolyPhen prediction and/or score"
    ),
    "MOTIF_NAME": AnnotationEntry(
        "MOTIF_NAME",
        description="The source and identifier of a transcription factor binding "
        "profile aligned at this position",
    ),
    "MOTIF_POS": AnnotationEntry(
        "MOTIF_POS",
        int,
        description="The relative position of the variation in the aligned TFBP",
    ),
    "HIGH_INF_POS": AnnotationEntry(
        "HIGH_INF_POS",
        typefunc="Y".__eq__,
        description="A flag indicating if the variant falls in a high information "
        "position of a transcription factor binding profile (TFBP)",
    ),
    "MOTIF_SCORE_CHANGE": AnnotationEntry(
        "MOTIF_SCORE_CHANGE",
        float32,
        description="The difference in motif score of the reference and variant "
        "sequences for the TFBP",
    ),
    "CELL_TYPE": AnnotationListEntry(
        "CELL_TYPE",
        ",",
        description="List of cell types and classifications for regulatory feature",
    ),
    "CANONICAL": AnnotationEntry(
        "CANONICAL",
        typefunc="YES".__eq__,
        description="A flag indicating if the transcript is denoted as the canonical "
        "transcript for this gene",
    ),
    "CCDS": AnnotationEntry(
        "CCDS", description="The CCDS identifer for this transcript, where applicable"
    ),
    "INTRON": AnnotationRangeTotalEntry(
        "INTRON", description="The intron number (out of total number)"
    ),
    "EXON": AnnotationRangeTotalEntry(
        "EXON", description="The exon index range (out of total number of exons)"
    ),
    "DOMAINS": AnnotationListDictEntry(
        "DOMAINS",
        description="The source and identifer of any overlapping protein domains",
    ),
    "DISTANCE": AnnotationEntry(
        "DISTANCE", int, description="Shortest distance from variant to transcript"
    ),
    "IND": AnnotationEntry("IND", description="Individual name"),
    # "ZYG": AnnotationEntry(
    #     "ZYG", description="Zygosity of individual genotype at this locus"
    # ),
    # "SV": AnnotationEntry("SV", description="IDs of overlapping structural variants"),
    # "FREQS": AnnotationEntry(
    #     "FREQS", description="Frequencies of overlapping variants used in filtering"
    # ),
    "AF": AnnotationEntry(
        "AF", float32, description="Frequency of existing variant in 1000 Genomes"
    ),
    "AFR_AF": AnnotationEntry(
        "AFR_AF",
        float32,
        description="Frequency of existing variant in 1000 Genomes combined "
        "African population",
    ),
    "AMR_AF": AnnotationEntry(
        "AMR_AF",
        float32,
        description="Frequency of existing variant in 1000 Genomes combined "
        "American population",
    ),
    "ASN_AF": AnnotationEntry(
        "ASN_AF",
        float32,
        description="Frequency of existing variant in 1000 Genomes combined "
        "Asian population",
    ),
    "EUR_AF": AnnotationEntry(
        "EUR_AF",
        float32,
        description="Frequency of existing variant in 1000 Genomes combined "
        "European population",
    ),
    "EAS_AF": AnnotationEntry(
        "EAS_AF",
        float32,
        description="Frequency of existing variant in 1000 Genomes combined "
        "East Asian population",
    ),
    "SAS_AF": AnnotationEntry(
        "SAS_AF",
        float32,
        description="Frequency of existing variant in 1000 Genomes combined "
        "South Asian population",
    ),
    "AA_AF": AnnotationEntry(
        "AA_AF",
        float32,
        description="Frequency of existing variant in NHLBI-ESP "
        "African American population",
    ),
    "EA_AF": AnnotationEntry(
        "EA_AF",
        float32,
        description="Frequency of existing variant in NHLBI-ESP "
        "European American population",
    ),
    "gnomAD_AF": AnnotationEntry(
        "gnomAD_AF",
        float32,
        description="Frequency of existing variant in gnomAD exomes "
        "combined population",
    ),
    "gnomAD_AFR_AF": AnnotationEntry(
        "gnomAD_AFR_AF",
        float32,
        description="Frequency of existing variant in gnomAD exomes "
        "African/American population",
    ),
    "gnomAD_AMR_AF": AnnotationEntry(
        "gnomAD_AMR_AF",
        float32,
        description="Frequency of existing variant in gnomAD exomes "
        "American population",
    ),
    "gnomAD_ASJ_AF": AnnotationEntry(
        "gnomAD_ASJ_AF",
        float32,
        description="Frequency of existing variant in gnomAD exomes "
        "Ashkenazi Jewish population",
    ),
    "gnomAD_EAS_AF": AnnotationEntry(
        "gnomAD_EAS_AF",
        float32,
        description="Frequency of existing variant in gnomAD exomes "
        "East Asian population",
    ),
    "gnomAD_FIN_AF": AnnotationEntry(
        "gnomAD_FIN_AF",
        float32,
        description="Frequency of existing variant in gnomAD exomes "
        "Finnish population",
    ),
    "gnomAD_NFE_AF": AnnotationEntry(
        "gnomAD_NFE_AF",
        float32,
        description="Frequency of existing variant in gnomAD exomes "
        "Non-Finnish European population",
    ),
    "gnomAD_OTH_AF": AnnotationEntry(
        "gnomAD_OTH_AF",
        float32,
        description="Frequency of existing variant in gnomAD exomes "
        "combined other combined populations",
    ),
    "gnomAD_SAS_AF": AnnotationEntry(
        "gnomAD_SAS_AF",
        float32,
        description="Frequency of existing variant in gnomAD exomes "
        "South Asian population",
    ),
    "MAX_AF": AnnotationEntry(
        "MAX_AF",
        float32,
        description="Maximum observed allele frequency in "
        "1000 Genomes, ESP and gnomAD",
    ),
    "MAX_AF_POPS": AnnotationListEntry(
        "MAX_AF_POPS",
        sep="&",
        description="Populations in which maximum allele frequency was observed",
    ),
    "CLIN_SIG": AnnotationListEntry(
        "CLIN_SIG",
        "&",
        description="ClinVar clinical significance of the dbSNP variant",
    ),
    "BIOTYPE": AnnotationEntry(
        "BIOTYPE", description="Biotype of transcript or regulatory feature"
    ),
    "APPRIS": AnnotationEntry(
        "APPRIS",
        description="Annotates alternatively spliced transcripts as primary or "
        "alternate based on a range of computational methods. "
        "NB: not available for GRCh37",
    ),
    "TSL": AnnotationEntry(
        "TSL", description="Transcript support level. NB: not available for GRCh37"
    ),
    "PUBMED": AnnotationListEntry(
        "PUBMED",
        description="Pubmed ID(s) of publications that cite existing variant",
        sep="&",
    ),
    "SOMATIC": AnnotationListEntry(
        "SOMATIC",
        description="Somatic status of existing variant(s); "
        "multiple values correspond to multiple values "
        "in the Existing_variation field",
        sep="&",
    ),
    "PHENO": AnnotationListEntry(
        "PHENO",
        sep="&",
        description="Indicates if existing variant is associated with a phenotype, "
        "disease or trait; "
        "multiple values correspond to multiple values "
        "in the Existing_variation field",
    ),
    # TODO list type with unconfirmed sep
    "GENE_PHENO": AnnotationListEntry(
        "GENE_PHENO",
        sep="&",
        description="Indicates if overlapped gene is associated with a phenotype, "
        "disease or trait",
    ),
    "ALLELE_NUM": AnnotationEntry(
        "ALLELE_NUM",
        int,
        description="Allele number from input; "
        "0 is reference, 1 is first alternate etc",
    ),
    # "MINIMISED": AnnotationEntry(
    #     "MINIMISED",
    #     description="Alleles in this variant have been converted to minimal "
    #                 "representation before consequence calculation",
    # ),
    # TODO parse flag:
    #  "indicates if this block of consequence data was picked by
    #  --flag_pick or --flag_pick_allele"
    # "PICK": AnnotationEntry(
    #     "PICK",
    #     description="Indicates if this block of consequence data was picked by "
    #                 "--flag_pick or --flag_pick_allele",
    # ),
    # "BAM_EDIT": AnnotationEntry(
    #     "BAM_EDIT", description="Indicates success or failure of edit using BAM file"
    # ),
    "GIVEN_REF": AnnotationEntry(
        "GIVEN_REF", description="Reference allele from input"
    ),
    "USED_REF": AnnotationEntry(
        "USED_REF", description="Reference allele as used to get consequences"
    ),
    # TODO parse flag:
    #  see https://www.ensembl.org/info/docs/tools/vep/vep_formats.html#other_fields
    # "REFSEQ_MATCH": AnnotationEntry(
    #     "REFSEQ_MATCH",
    #     description="The RefSeq transcript match status; contains a number of flags "
    #                 "indicating whether this RefSeq transcript matches "
    #                 "the underlying reference sequence and/or an Ensembl transcript",
    # ),
    "OverlapBP": AnnotationEntry(
        "OverlapBP",
        int,
        description="Number of base pairs overlapping with the corresponding "
        "structural variation feature",
    ),
    "OverlapPC": AnnotationEntry(
        "OverlapPC",
        float32,
        description="Percentage of corresponding structural variation feature "
        "overlapped by the given input",
    ),
    # "CHECK_REF": AnnotationEntry(
    #     "CHECK_REF",
    #     description="Reports variants where the input reference does not match "
    #                 "the expected reference",
    # ),
    "AMBIGUITY": AnnotationEntry(
        "AMBIGUITY", description="IUPAC allele ambiguity code"
    ),
    "Amino_acids": AnnotationListEntry(
        "Amino_acids", description="Reference and variant amino acids", sep="/"
    ),
    "Codons": AnnotationListEntry(
        "Codons", description="Reference and variant codon sequence", sep="/"
    ),
    # TODO HGNC_ID description
    "HGNC_ID": AnnotationEntry("HGNC_ID"),
    "MANE": AnnotationEntry(
        "MANE", description="Matched Annotation from NCBI and EMBL-EBI (MANE)."
    ),
    "MANE_SELECT": AnnotationEntry(
        "MANE_SELECT",
        description="Matched Annotation from NCBI and EMBL-EBI (MANE) "
        "canonical transcript, indicated by the respective RefSeq NM ID."
        "For more info, see: "
        "https://www.ensembl.org/info/genome/genebuild/mane.html",
    ),
    "MANE_PLUS_CLINICAL": AnnotationEntry(
        "MANE_PLUS_CLINICAL",
        description="Matched Annotation from NCBI and EMBL-EBI (MANE) "
        "transcripts beyond MANE_SELECT, that are important clinically. "
        "Indicated by their RefSeq NM IDs."
        "For more info, see: "
        "https://www.ensembl.org/info/genome/genebuild/mane.html",
    ),
    "GO": AnnotationEntry("GO", description="Gene ontology (GO) terms."),
    "miRNA": AnnotationEntry(
        "miRNA",
        description="Determines where in the secondary structure of a miRNA "
        "a variant falls",
    ),
    "Existing_variation": AnnotationListEntry(
        "Existing_variation",
        sep="&",
        description="Identifier(s) of co-located known variants",
    ),
    "LoFtool": AnnotationEntry(
        "LoFtool",
        typefunc=float32,
        description="Provides a rank of genic intolerance and "
        "consequent susceptibility to disease based on the ratio of Loss-of-function "
        "(LoF) to synonymous mutations.",
    ),
    "REVEL": AnnotationEntry(
        "REVEL",
        typefunc=float32,
        description="Estimate of the pathogenicity of missense variants. "
        "Please cite the REVEL publication alongside the VEP if you use this "
        "resource: https://www.ncbi.nlm.nih.gov/pubmed/27666373",
    ),
    "ExACpLI": AnnotationEntry(
        "ExACpLI",
        typefunc=float32,
        description="Probabililty of a gene being loss-of-function intolerant (pLI). "
        "Please cite the respective ExAC publication alongside the VEP if you use this "
        "resource: https://www.ncbi.nlm.nih.gov/pubmed/27535533",
    ),
}

# VEP now uses gnomADe_* and gnomADg_* instead of gnomAD_*.
# Therefore, we need to add these new entries to the KNOWN_ANN_TYPE_MAP_VEP
NEW_GNOMAD_ENTRIES = {}
for key, typer in KNOWN_ANN_TYPE_MAP_VEP.items():
    if key.startswith("gnomAD_"):
        sub_key = key.split("_", 1)[1]
        for t in ("e", "g"):
            name = f"gnomAD{t}_{sub_key}"
            description = typer.description()
            if t == "g":
                description = description.replace("gnomAD exomes", "gnomAD genomes")
            typer = AnnotationEntry(
                name=name,
                typefunc=typer._typefunc,
                description=description,
            )
            NEW_GNOMAD_ENTRIES[name] = typer
KNOWN_ANN_TYPE_MAP_VEP.update(NEW_GNOMAD_ENTRIES)
KNOWN_ANN_TYPE_MAP = {**KNOWN_ANN_TYPE_MAP_SNPEFF, **KNOWN_ANN_TYPE_MAP_VEP}

ANN_TYPER = AnnotationTyper(KNOWN_ANN_TYPE_MAP)



import argparse

from . import __version__
from .modules import annotate, filter, table, tag


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {__version__}"
    )
    subparsers = parser.add_subparsers(
        dest="command", description="valid subcommands", required=True
    )
    filter.add_subcommmand(subparsers)
    table.add_subcommmand(subparsers)
    annotate.add_subcommmand(subparsers)
    tag.add_subcommand(subparsers)

    args = parser.parse_args()
    if args.command == "filter":
        filter.execute(args)
    elif args.command == "table":
        table.execute(args)
    elif args.command == "annotate":
        annotate.execute(args)
    elif args.command == "tag":
        tag.execute(args)
    else:
        raise ValueError(f"Unknown subcommand {args.command}")



import argparse
import ast
import shlex
from typing import Dict, Iterable, Iterator, List, Optional, Set

from pysam.libcbcf import VariantHeader, VariantRecord

from .errors import InvalidExpression


def check_expression(expression: str) -> str:
    if ".__" in expression:
        raise InvalidExpression(expression, "The expression must not contain '.__'")
    try:
        tree = ast.parse(expression, mode="eval")
        if isinstance(tree.body, (ast.BoolOp, ast.Compare)):
            return expression
        else:
            # TODO possibly check for ast.Call, func return type
            return expression
    except SyntaxError:
        raise InvalidExpression(
            expression, "The expression has to be syntactically correct."
        )


def swap_quotes(s: str) -> str:
    return s.replace('"', '\\"').replace("'", '"').replace('\\"', "'")


def single_outer(s: str) -> bool:
    if '"' in s and "'" in s:
        return s.index('"') > s.index("'")
    elif '"' in s:
        return True
    elif "'" in s:
        return False
    return True


def normalize(s: str) -> str:
    return shlex.quote(swap_quotes(s) if not single_outer(s) else s)


def get_annotation_keys(header: VariantHeader, ann_key: str) -> List[str]:
    separator = "'"
    for rec in header.records:
        if rec.key == "VEP":
            separator = ":"
            continue
        if rec.get("ID") == ann_key:
            return list(
                map(
                    str.strip,
                    rec.get("Description").strip('"').split(separator)[1].split("|"),
                )
            )
    return []


def split_annotation_entry(entry: str) -> List[str]:
    return entry.split("|")


def is_bnd_record(record: VariantRecord) -> bool:
    return "SVTYPE" in record.info and record.info.get("SVTYPE", None) == "BND"


class BreakendEvent(object):
    __slots__ = ["name", "keep", "records", "keep_records", "mate_pair"]

    def __init__(self, name: str, mate_pair: bool = False):
        self.name = name
        self.records: List[VariantRecord] = []
        self.keep_records: List[bool] = []
        self.keep = False
        self.mate_pair = mate_pair

    def add(self, record: VariantRecord, keep_record: bool):
        self.records.append(record)
        self.keep_records.append(keep_record)
        self.keep |= keep_record

    def emit(self) -> Iterator[VariantRecord]:
        assert self.keep
        yield from self.records
        self.records = []
        self.keep_records = []
        # do not reset self.keep!

    def is_mate_pair(self) -> bool:
        return self.mate_pair

    def __str__(self):
        return self.name

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return self.name == other.name


def mate_key(mates: Iterable[Optional[str]]) -> str:
    return "__MATES: " + ",".join(sorted(m for m in mates if m is not None))


class AppendTagExpression(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        assert len(values) == 1
        if not hasattr(namespace, self.dest) or getattr(namespace, self.dest) is None:
            setattr(namespace, self.dest, dict())
        value = values[0].strip()
        key, value = value.strip().split("=", 1)
        expr = check_expression(value)
        getattr(namespace, self.dest)[key.strip()] = expr


class AppendKeyValuePair(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        assert len(values) == 1
        if not hasattr(namespace, self.dest) or getattr(namespace, self.dest) is None:
            setattr(namespace, self.dest, dict())
        value = values[0].strip()
        key, value = value.strip().split("=", 1)
        getattr(namespace, self.dest)[key.strip()] = value


def read_auxiliary(aux: Dict[str, str]) -> Dict[str, Set[str]]:
    # read auxiliary files, split at any whitespace and store contents in a set
    def read_set(path: str) -> Set[str]:
        with open(path, "rt") as f:
            return set(line.rstrip() for line in f)

    return {name: read_set(contents) for name, contents in aux.items()}



import re
import sys
from itertools import chain, islice
from sys import stderr
from typing import Dict, Iterator, Set, Tuple

from pysam.libcbcf import VariantFile, VariantHeader, VariantRecord

from .. import __version__
from ..common import (
    AppendKeyValuePair,
    AppendTagExpression,
    check_expression,
    normalize,
    read_auxiliary,
    single_outer,
    swap_quotes,
)
from ..errors import FilterAlreadyDefined, FilterTagNameInvalid, VembraneError
from ..representations import Environment
from .filter import DeprecatedAction


def add_subcommand(subparsers):
    parser = subparsers.add_parser("tag")
    parser.register("action", "deprecated", DeprecatedAction)
    parser.add_argument(
        "--tag",
        "-t",
        help="Tag records using the FILTER field.\n"
        "Note: tag names cannot contain `;` or whitespace and must not be '0'.\n"
        'Example: `--tag q_above_30="not (QUAL<=30)"`',
        nargs=1,
        metavar="TAG=EXPRESSION",
        action=AppendTagExpression,
        required=True,
    )
    parser.add_argument(
        "vcf", help="The file containing the variants.", nargs="?", default="-"
    )
    parser.add_argument(
        "--output",
        "-o",
        default="-",
        help="Output file, if not specified, output is written to STDOUT.",
    )
    parser.add_argument(
        "--output-fmt",
        "-O",
        default="vcf",
        choices=["vcf", "bcf", "uncompressed-bcf"],
        help="Output format.",
    )
    parser.add_argument(
        "--annotation-key",
        "-k",
        metavar="FIELDNAME",
        default="ANN",
        help="The INFO key for the annotation field.",
    )
    parser.add_argument(
        "--aux",
        "-a",
        nargs=1,
        action=AppendKeyValuePair,
        metavar="NAME=PATH",
        default={},
        help="Path to an auxiliary file containing a set of symbols",
    )
    parser.add_argument(
        "--overwrite-number-info",
        nargs=2,
        action="append",
        metavar=("FIELD", "NUMBER"),
        default=[],
        help="Overwrite the number specification for INFO fields "
        "given in the VCF header. "
        "Example: `--overwrite-number cosmic_CNT=.`",
    )
    parser.add_argument(
        "--overwrite-number-format",
        nargs=2,
        action="append",
        metavar=("FIELD", "NUMBER"),
        default=[],
        help="Overwrite the number specification for FORMAT fields "
        "given in the VCF header. "
        "Example: `--overwrite-number-format DP=2`",
    )
    parser.add_argument(
        "--tag-mode",
        "-m",
        default="pass",
        choices=["pass", "fail"],
        help="Set, whether to tag records that pass the tag expression(s), "
        "or records that fail them."
        "By default, vembrane tags records for which the tag expression(s) pass. "
        "This allows for descriptive tag names such as `q_at_least_30`, "
        "which would correspond to an expression `QUAL >= 30`. "
        "However, the VCF specification (`v4.4`) defines tags to be set when a "
        "filter expression is failed, so vembrane also offers the `fail` mode.",
    )


def test_record(
    env: Environment,
    idx: int,
    record: VariantRecord,
    ann_key: str,
) -> Tuple[VariantRecord, bool]:
    env.update_from_record(idx, record)
    if env.expression_annotations():
        # if the expression contains a reference to the ANN field
        # get all annotations from the record.info field
        # (or supply an empty ANN value if the record has no ANN field)
        try:
            annotations = record.info[ann_key]
        except KeyError:
            annotations = [""]

        #  … and check if the expression evaluates to true for any  of the annotations
        filtered = any(map(env.evaluate, annotations))
        return record, filtered
    else:
        # otherwise, the annotations are irrelevant w.r.t. the expression,
        # so we can omit them
        return record, env.evaluate()


def tag_vcf(
    vcf: VariantFile,
    expressions: Dict[str, str],
    ann_key: str,
    auxiliary: Dict[str, Set[str]] = {},
    overwrite_number: Dict[str, Dict[str, str]] = {},
    invert: bool = False,
) -> Iterator[VariantRecord]:
    # For each tag-expression pair, a different Environment must be used.
    envs = {
        tag: Environment(expression, ann_key, vcf.header, auxiliary, overwrite_number)
        for tag, expression in expressions.items()
    }

    record: VariantRecord
    for idx, record in enumerate(vcf):
        for tag, env in envs.items():
            record, keep = test_record(env, idx, record, ann_key)
            if invert:
                keep = not keep
            if keep:
                record.filter.add(tag)
        yield record


def check_tag(tag: str):
    if re.search(r"^0$|[\s;]", tag):
        raise FilterTagNameInvalid(tag)


def execute(args) -> None:
    aux = read_auxiliary(args.aux)
    with VariantFile(args.vcf) as vcf:
        header: VariantHeader = vcf.header

        overwrite_number = {
            "INFO": dict(args.overwrite_number_info),
            "FORMAT": dict(args.overwrite_number_format),
        }

        expressions = dict(args.tag)
        for tag, expr in expressions.items():
            for t, rec in vcf.header.filters.items():
                if t == tag:
                    e = FilterAlreadyDefined(tag)
                    print(e, file=stderr)
                    exit(1)
            try:
                check_tag(tag)
            except VembraneError as ve:
                print(ve, file=stderr)
                exit(1)
            expr = swap_quotes(expr) if single_outer(expr) else expr
            check_expression(expr)
            vcf.header.add_meta(
                key="FILTER", items=[("ID", tag), ("Description", expr)]
            )

        header.add_meta("vembraneVersion", __version__)
        header.add_meta(
            "vembraneCmd",
            "vembrane "
            + " ".join(normalize(arg) if " " in arg else arg for arg in sys.argv[1:]),
        )

        records = tag_vcf(
            vcf,
            expressions,
            args.annotation_key,
            auxiliary=aux,
            overwrite_number=overwrite_number,
            invert=(args.tag_mode == "fail"),
        )

        try:
            first_record = list(islice(records, 1))
        except VembraneError as ve:
            print(ve, file=stderr)
            exit(1)

        records = chain(first_record, records)
        fmt = {"vcf": "", "bcf": "b", "uncompressed-bcf": "u"}[args.output_fmt]
        with VariantFile(
            args.output,
            f"w{fmt}",
            header=header,
        ) as out:
            try:
                for record in records:
                    out.write(record)

            except VembraneError as ve:
                print(ve, file=stderr)
                exit(1)



import argparse
import sys
from collections import defaultdict
from itertools import chain, islice
from sys import stderr
from typing import Dict, Iterator, List, Optional, Set, Tuple

import yaml
from pysam.libcbcf import VariantFile, VariantHeader, VariantRecord

from .. import __version__
from ..common import (
    AppendKeyValuePair,
    BreakendEvent,
    check_expression,
    get_annotation_keys,
    is_bnd_record,
    mate_key,
    normalize,
    read_auxiliary,
    split_annotation_entry,
)
from ..errors import VembraneError
from ..representations import Environment


class DeprecatedAction(argparse.Action):
    def __call__(self, *args, **kwargs):
        print(
            f"{'/'.join(self.option_strings)} is deprecated.\n{self.help}",
            file=sys.stderr,
        )
        exit(1)


def add_subcommmand(subparsers):
    parser = subparsers.add_parser("filter")
    parser.register("action", "deprecated", DeprecatedAction)
    parser.add_argument(
        "expression",
        type=check_expression,
        help="Filter variants and annotations. If this removes all annotations, "
        "the variant is removed as well.",
    )
    parser.add_argument(
        "vcf", help="The file containing the variants.", nargs="?", default="-"
    )
    parser.add_argument(
        "--output",
        "-o",
        default="-",
        help="Output file, if not specified, output is written to STDOUT.",
    )
    parser.add_argument(
        "--output-fmt",
        "-O",
        default="vcf",
        choices=["vcf", "bcf", "uncompressed-bcf"],
        help="Output format.",
    )
    parser.add_argument(
        "--annotation-key",
        "-k",
        metavar="FIELDNAME",
        default="ANN",
        help="The INFO key for the annotation field.",
    )
    parser.add_argument(
        "--aux",
        "-a",
        nargs=1,
        action=AppendKeyValuePair,
        default={},
        metavar="NAME=PATH",
        help="Path to an auxiliary file containing a set of symbols",
    )
    parser.add_argument(
        "--statistics",
        "-s",
        metavar="FILE",
        default=None,
        help="Write statistics to this file.",
    )
    parser.add_argument(
        "--keep-unmatched",
        default=False,
        action="store_true",
        help="Keep all annotations of a variant if at least one of them passes "
        "the expression.",
    )
    parser.add_argument(
        "--preserve-order",
        default=False,
        action="store_true",
        help="Ensures that the order of the output matches that of the \
              input. This is only useful if the input contains breakends (BNDs) \
              since the order of all other variants is preserved anyway.",
    )
    parser.add_argument(
        "--overwrite-number-info",
        nargs=1,
        action=AppendKeyValuePair,
        metavar="FIELD=NUMBER",
        default={},
        help="Overwrite the number specification for INFO fields "
        "given in the VCF header. "
        "Example: `--overwrite-number cosmic_CNT=.`",
    )
    parser.add_argument(
        "--overwrite-number-format",
        nargs=1,
        action=AppendKeyValuePair,
        metavar="FIELD=NUMBER",
        default={},
        help="Overwrite the number specification for FORMAT fields "
        "given in the VCF header. "
        "Example: `--overwrite-number-format DP=2`",
    )


def test_and_update_record(
    env: Environment,
    idx: int,
    record: VariantRecord,
    ann_key: str,
    keep_unmatched: bool,
) -> Tuple[VariantRecord, bool]:
    env.update_from_record(idx, record)
    if env.expression_annotations():
        # if the expression contains a reference to the ANN field
        # get all annotations from the record.info field
        # (or supply an empty ANN value if the record has no ANN field)
        try:
            annotations = record.info[ann_key]
        except KeyError:
            num_ann_entries = len(env._annotation._ann_conv.keys())
            empty = "|" * num_ann_entries
            print(
                f"No ANN field found in record {idx}, "
                f"replacing with NAs (i.e. 'ANN={empty}')",
                file=sys.stderr,
            )
            annotations = [empty]

        #  … and only keep the annotations where the expression evaluates to true
        if keep_unmatched:
            filtered = any(map(env.evaluate, annotations))
            return record, filtered
        else:
            filtered_annotations = [
                annotation for annotation in annotations if env.evaluate(annotation)
            ]

            if len(annotations) != len(filtered_annotations):
                # update annotations if they have actually been filtered
                record.info[ann_key] = filtered_annotations

            return record, len(filtered_annotations) > 0
    else:
        # otherwise, the annotations are irrelevant w.r.t. the expression,
        # so we can omit them
        return record, env.evaluate()


def filter_vcf(
    vcf: VariantFile,
    expression: str,
    ann_key: str,
    keep_unmatched: bool = False,
    preserve_order: bool = False,
    auxiliary: Dict[str, Set[str]] = {},
    overwrite_number: Dict[str, Dict[str, str]] = {},
) -> Iterator[VariantRecord]:
    env = Environment(expression, ann_key, vcf.header, auxiliary, overwrite_number)
    has_mateid_key = vcf.header.info.get("MATEID", None) is not None
    has_event_key = vcf.header.info.get("EVENT", None) is not None

    def get_event_name(
        record: VariantRecord,
        has_mateid_key=has_mateid_key,
        has_event_key=has_event_key,
    ) -> Tuple[Optional[str], Optional[str]]:
        mate_ids: List[str] = record.info.get("MATEID", []) if has_mateid_key else []
        event_name: Optional[str] = (
            record.info.get("EVENT", None) if has_event_key else None
        )

        if len(mate_ids) > 1 and not event_name:
            raise ValueError(
                f"Filtering of BND records with multiple mates is unsupported "
                f"(see VCF 4.3, section 5.4.3 'Multiple mates'):\n{str(record)}"
            )

        mate_id = mate_ids[0] if len(mate_ids) == 1 else None
        mate_pair = mate_key([record.id, mate_id]) if mate_ids else None

        return event_name, mate_pair

    record: VariantRecord
    if not preserve_order:
        # If the order is not important, emit records that pass the filter expression
        # as we encounter them
        # However, breakends have to be considered jointly, so keep track of the
        # respective events.
        event_dict: Dict[str, BreakendEvent] = dict()
        for idx, record in enumerate(vcf):
            record, keep = test_and_update_record(
                env, idx, record, ann_key, keep_unmatched
            )

            # Breakend records *may* have the "EVENT" specified, but don't have to.
            # In that case only the MATEID *may* be available
            # (which may contain more than one ID)
            if is_bnd_record(record):
                event_name, mate_pair_name = get_event_name(record)

                # if EVENT is set, it has priority over MATEID.
                event_name = event_name or mate_pair_name

                # if both EVENT and MATEID are not set
                # we can only treat this BND record as a regular one,
                # (or somehow try to guess its correct mate)
                if not event_name:
                    print(
                        f"Warning: Encountered breakend record at index {idx} "
                        f"without either of MATEID or EVENT specified, "
                        f"treating it as a regular record:\n{str(record)}",
                        file=stderr,
                    )
                    if keep:
                        yield record
                    continue

                event = event_dict.get(event_name, None)

                # if there's already an associated event
                if event:
                    # add this record to the event
                    event.add(record, keep)

                    # if we already know that the event is a "PASS"…
                    if event.keep:
                        # … emit all records associated with it
                        yield from event.emit()

                        # in the case of a simple mate pair, we can delete the event
                        # at this point, because no more records will be added to it
                        if event.is_mate_pair():
                            del event_dict[mate_pair_name]
                else:
                    # if there's no entry for the event or mate pair yet, create one
                    is_mate_pair = mate_pair_name and mate_pair_name == event_name
                    event = BreakendEvent(event_name, is_mate_pair)
                    event.add(record, keep)
                    event_dict[event_name] = event
            elif keep:
                yield record

        if len(event_dict) > 0:
            # output BNDs if any are left unprocessed
            for event_name, event in event_dict.items():
                if event_name and event.keep:
                    yield from event.emit()
    elif preserve_order:
        # If order *is* important, the first pass cannot emit any records but only
        # keep track of breakend events. The records will only be emitted during the
        # second pass.
        event_set: Set[str] = set()

        def fallback_name(record: VariantRecord) -> str:
            event_name, mate_pair_name = get_event_name(record)

            # There may be BND record which have neither of
            # INFO['EVENT'], INFO['MATEID'] or even ID specified.
            # In that case, we still need a pseudo-event name,
            # which we construct from the record's index in the file.
            return event_name or mate_pair_name or record.id or f"DUMMY: {idx}"

        for idx, record in enumerate(vcf):
            if is_bnd_record(record):
                record, keep = test_and_update_record(
                    env, idx, record, ann_key, keep_unmatched
                )
                if keep:
                    event_name = fallback_name(record)
                    event_set.add(event_name)

        # The second pass can now yield records in the correct order
        vcf.reset()
        for idx, record in enumerate(vcf):
            if is_bnd_record(record):
                event_name = fallback_name(record)
                if event_name in event_set:
                    yield record
            else:
                record, keep = test_and_update_record(
                    env, idx, record, ann_key, keep_unmatched
                )
                if keep:
                    yield record


def statistics(
    records: Iterator[VariantRecord], vcf: VariantFile, filename: str, ann_key: str
) -> Iterator[VariantRecord]:
    annotation_keys = get_annotation_keys(vcf.header, ann_key)
    counter = defaultdict(lambda: defaultdict(lambda: 0))
    for record in records:
        for annotation in record.info[ann_key]:
            for key, raw_value in zip(
                annotation_keys, split_annotation_entry(annotation)
            ):
                value = raw_value.strip()
                if value:
                    counter[key][value] += 1
        yield record

    # reduce dicts with many items, to just one counter
    for key, subdict in counter.items():
        if len(subdict) > 10:
            counter[key] = f"#{len(subdict)}"

    yaml.add_representer(defaultdict, yaml.representer.Representer.represent_dict)
    with open(filename, "w") as out:
        yaml.dump(dict(counter), out)


def execute(args) -> None:
    aux = read_auxiliary(args.aux)
    with VariantFile(args.vcf) as vcf:
        header: VariantHeader = vcf.header
        header.add_meta("vembraneVersion", __version__)
        # NOTE: If .modules.filter.execute might be used as a library function
        #       in the future, we should not record sys.argv directly below.
        cmd_parts = [normalize(arg) if " " in arg else arg for arg in sys.argv[3:]]
        expr = " ".join(a.strip() for a in args.expression.split("\n"))
        expr = normalize(expr)

        header.add_meta(
            "vembraneCmd",
            "vembrane " + expr + " " + " ".join(cmd_parts),
        )

        overwrite_number = {
            "INFO": dict(args.overwrite_number_info),
            "FORMAT": dict(args.overwrite_number_format),
        }

        records = filter_vcf(
            vcf,
            args.expression,
            args.annotation_key,
            keep_unmatched=args.keep_unmatched,
            preserve_order=args.preserve_order,
            auxiliary=aux,
            overwrite_number=overwrite_number,
        )

        try:
            first_record = list(islice(records, 1))
        except VembraneError as ve:
            print(ve, file=stderr)
            exit(1)

        records = chain(first_record, records)

        if args.statistics is not None:
            records = statistics(records, vcf, args.statistics, args.annotation_key)

        fmt = {"vcf": "", "bcf": "b", "uncompressed-bcf": "u"}[args.output_fmt]
        with VariantFile(
            args.output,
            f"w{fmt}",
            header=header,
        ) as out:
            try:
                for record in records:
                    out.write(record)

            except VembraneError as ve:
                print(ve, file=stderr)
                exit(1)



from collections.abc import Iterable
from sys import stderr
from typing import Any, Callable, Dict, Iterator

import numpy as np
import yaml
from intervaltree import Interval, IntervalTree
from pysam.libcbcf import VariantFile, VariantRecord

from ..common import check_expression
from ..representations import Environment


def add_subcommmand(subparsers):
    parser = subparsers.add_parser("annotate")
    parser.add_argument(
        "config",
        type=check_expression,
        help="The configuration file.",
    )
    parser.add_argument(
        "vcf", help="The file containing the variants.", nargs="?", default="-"
    )
    parser.add_argument(
        "--annotation-key",
        "-k",
        metavar="FIELDNAME",
        default="ANN",
        help="The INFO key for the annotation field.",
    )
    parser.add_argument(
        "--output",
        "-o",
        default="-",
        help="Output file, if not specified, output is written to STDOUT.",
    )
    parser.add_argument(
        "--output-fmt",
        "-O",
        default="vcf",
        choices=["vcf", "bcf", "uncompressed-bcf"],
        help="Output format.",
    )


typeparser: Dict[str, Callable[[str], Any]] = {
    "Float": float,
    "Integer": int,
    "Character": lambda x: str(x)[0],
    "String": str,
    "Flag": bool,
}


def annotate_vcf(
    vcf: VariantFile,
    expression: str,
    ann_key: str,
    ann_data,
    config: dict,
) -> Iterator[VariantRecord]:
    env = Environment(expression, ann_key, vcf.header)

    config_chrom_column = config["annotation"]["columns"]["chrom"]
    config_start_column = config["annotation"]["columns"]["start"]
    config_stop_column = config["annotation"]["columns"]["stop"]

    available_chromsomes = set(np.unique(ann_data[config_chrom_column]))

    tree = dict()
    chrom_ann_data = dict()
    for chrom in available_chromsomes:
        d = ann_data[ann_data[config_chrom_column] == chrom]
        chrom_ann_data[chrom] = d
        tree[chrom] = IntervalTree(
            Interval(d[config_start_column], d[config_stop_column], i)
            for i, d in enumerate(d)
        )

        # tree = IntervalTree.from_tuples(interval_tuples))

    current_chrom = None
    current_ann_data = None

    record: VariantRecord
    for idx, record in enumerate(vcf):
        chrom = None
        if current_chrom != record.chrom:
            current_chrom = record.chrom

            # find the correct chrom name
            tmp = current_chrom
            if tmp.lower().startswith("chr"):
                tmp = tmp[3:]
            for prefix in ["", "chr", "Chr", "CHR"]:
                if prefix + tmp in available_chromsomes:
                    chrom = prefix + tmp
                    current_ann_data = chrom_ann_data[chrom]
                    t = tree[chrom]

        if chrom:
            indices = np.fromiter((i for _, _, i in t[record.start]), dtype=int)
            if len(indices):
                env.update_data(current_ann_data[(np.array(indices))])
                env.update_from_record(idx, record)
                ann_values = env.table()

                for v, expression_value in zip(
                    map(lambda x: x["value"], config["annotation"]["values"]),
                    ann_values,
                ):
                    if not v["number"] == ".":
                        number = int(v["number"])
                    else:
                        number = -1

                    parse = typeparser[v["type"]]
                    if number == -1:
                        expression_value = list(map(parse, expression_value))
                    elif number > 1:
                        expression_value = list(map(parse, expression_value))
                        assert len(expression_value) == number
                    else:
                        # number == 1
                        assert isinstance(expression_value, str) or not isinstance(
                            expression_value, Iterable
                        )
                        expression_value = parse(expression_value)
                    record.info[v["vcf_name"]] = expression_value

        yield record


def execute(args):
    with open(args.config, "r") as stream:
        try:
            config = yaml.safe_load(stream)
        except yaml.YAMLError as e:
            print(e, file=stderr)
            exit(1)

    # load annotation data
    ann_data = np.genfromtxt(
        config["annotation"]["file"],
        delimiter=config["annotation"].get("delimiter", "\t"),
        names=True,
        dtype=None,
        encoding=None,
    )
    # ann_data = pd.read_csv(config["annotation"]["file"], sep="\t", header=0)
    # ann_data = dict(tuple(ann_data.groupby("chrom")))

    # build expression
    expression = ",".join(
        f'{value["expression"]}'
        for value in map(lambda x: x["value"], config["annotation"]["values"])
    )
    expression = f"({expression})"

    with VariantFile(args.vcf) as vcf:
        # add new info
        for value in config["annotation"]["values"]:
            value = value["value"]
            vcf.header.add_meta(
                "INFO",
                items=[
                    ("ID", value["vcf_name"]),
                    ("Number", value["number"]),
                    ("Type", value["type"]),
                    ("Description", value["description"]),
                ],
            )

        fmt = {"vcf": "", "bcf": "b", "uncompressed-bcf": "u"}[args.output_fmt]
        with VariantFile(
            args.output,
            f"w{fmt}",
            header=vcf.header,
        ) as out:
            variants = annotate_vcf(
                vcf,
                expression,
                args.annotation_key,
                ann_data=ann_data,
                config=config,
            )
            for v in variants:
                out.write(v)



import contextlib
import csv
import sys
from sys import stderr
from typing import Any, Dict, Iterator, List, Set

import asttokens
from pysam.libcbcf import VariantFile, VariantRecord

from ..common import AppendKeyValuePair, check_expression, read_auxiliary
from ..errors import HeaderWrongColumnNumber, VembraneError
from ..globals import allowed_globals
from ..representations import Environment
from .filter import DeprecatedAction


def add_subcommmand(subparsers):
    parser = subparsers.add_parser("table")
    parser.register("action", "deprecated", DeprecatedAction)
    parser.add_argument(
        "expression",
        type=check_expression,
        help="The expression for the output.",
    )
    parser.add_argument(
        "vcf", help="The file containing the variants.", nargs="?", default="-"
    )
    parser.add_argument(
        "--annotation-key",
        "-k",
        metavar="FIELDNAME",
        default="ANN",
        help="The INFO key for the annotation field.",
    )
    parser.add_argument(
        "--separator",
        "-s",
        default="\t",
        metavar="CHAR",
        help="Define the field separator (default: \\t).",
    )
    parser.add_argument(
        "--header",
        default="auto",
        metavar="TEXT",
        help='Override the automatically generated header. Provide "auto" (default) \
              to automatically generate the header from the expression. Provide a \
              comma separated string to manually set the header. Provide "none" to \
              disable any header output.',
    )
    parser.add_argument(
        "--long",
        help="Instead of using `for_each_sample` to generate multiple columns "
        "(wide format), "
        "use only one sample column but one row for each sample (long format).",
        action="store_true",
    )
    parser.add_argument(
        "--output",
        "-o",
        default="-",
        help="Output file, if not specified, output is written to STDOUT.",
    )
    parser.add_argument(
        "--aux",
        "-a",
        nargs=1,
        action=AppendKeyValuePair,
        metavar="NAME=PATH",
        default={},
        help="Path to an auxiliary file containing a set of symbols",
    )
    parser.add_argument(
        "--overwrite-number-info",
        nargs=1,
        action=AppendKeyValuePair,
        metavar="FIELD=NUMBER",
        default={},
        help="Overwrite the number specification for INFO fields "
        "given in the VCF header. "
        "Example: `--overwrite-number cosmic_CNT=.`",
    )
    parser.add_argument(
        "--overwrite-number-format",
        nargs=1,
        action=AppendKeyValuePair,
        metavar="FIELD=NUMBER",
        default={},
        help="Overwrite the number specification for FORMAT fields "
        "given in the VCF header. "
        "Example: `--overwrite-number-format DP=2`",
    )


def tableize_vcf(
    vcf: VariantFile,
    expression: str,
    ann_key: str,
    overwrite_number: Dict[str, Dict[str, str]] = {},
    long: bool = False,
    auxiliary: Dict[str, Set[str]] = {},
) -> Iterator[tuple]:
    kwargs: Dict[str, Any] = dict(
        overwrite_number=overwrite_number, auxiliary=auxiliary
    )
    if long:
        kwargs[
            "evaluation_function_template"
        ] = "lambda: (({expression}) for SAMPLE in SAMPLES)"
    else:
        expression = f"({expression})"
    env = Environment(expression, ann_key, vcf.header, **kwargs)

    record: VariantRecord
    for idx, record in enumerate(vcf):
        env.update_from_record(idx, record)
        if env.expression_annotations():
            # if the expression contains a reference to the ANN field
            # get all annotations from the record.info field
            # (or supply an empty ANN value if the record has no ANN field)
            try:
                annotations = record.info[ann_key]
            except KeyError:
                num_ann_entries = len(env._annotation._ann_conv.keys())
                empty = "|" * num_ann_entries
                print(
                    f"No ANN field found in record {idx}, "
                    f"replacing with NAs (i.e. 'ANN={empty}')",
                    file=sys.stderr,
                )
                annotations = [empty]
            for annotation in annotations:
                if long:
                    yield from env.table(annotation)
                else:
                    yield env.table(annotation)
        else:
            if long:
                yield from env.table()
            else:
                yield env.table()


def generate_for_each_sample_expressions(s: str, vcf: VariantFile) -> List[str]:
    from asttokens.util import replace

    # parse the `for_each_sample(lambda var: inner) expression
    var, inner = _var_and_body(s)

    # here we parse the inner part again, so that position offsets start at 0
    tok = asttokens.ASTTokens(inner, parse=True)

    # *replace* `var` with each value from the list of samples
    samples = list(vcf.header.samples)
    expanded = []
    for sample in samples:
        replacements = []
        for t in tok.tokens:
            if t.type == 1 and t.string == var:
                pos = (t.startpos, t.endpos)
                replacements.append((*pos, f"'{sample}'"))
        expanded.append(replace(inner, replacements))
    return expanded


def generate_for_each_sample_column_names(s: str, vcf: VariantFile) -> List[str]:
    # parse the `for_each_sample(lambda var: inner) expression
    var, inner = _var_and_body(s)

    samples = list(vcf.header.samples)
    __globals = allowed_globals.copy()

    column_names = []
    for sample in samples:
        __globals[var] = sample
        column_name = eval(inner, __globals, {})
        if not isinstance(column_name, (str, bytes)):
            if hasattr(column_name, "__str__"):
                column_name = str(column_name)
            else:
                raise ValueError(
                    "The specified header expression does not evaluate to a string."
                    "Consider using `str(expression)` instead."
                )
        column_names.append(column_name)
    return column_names


def _var_and_body(s):
    import ast

    tok = asttokens.ASTTokens(s, parse=True)
    tok.mark_tokens(tok.tree)
    var = None
    inner = ""

    # for_each_sample must be the top level function
    # since that is the only place it is allowed in order to split the header expression
    # such that there is one column per sample (instead of *one* list-valued column)

    # walk the resulting AST, find the "for_each_sample" ast.Call node,
    # and extract the variable name and lambda body code from that
    for node in asttokens.util.walk(tok.tree):
        if isinstance(node, ast.Call):
            if hasattr(node.func, "id"):
                if node.func.id == "for_each_sample":
                    for arg in node.args:
                        if isinstance(arg, ast.Lambda):
                            var = tok.get_text(arg.args)
                            inner = tok.get_text(arg.body)
    return var, inner


def preprocess_expression(
    header: str, vcf: VariantFile, make_expression: bool = True
) -> str:
    """
    Split the header expression at toplevel commas into parts.
    Then, if one of these parts starts with 'for_each_sample',
    that part is expanded for each sample in vcf.header.samples
    """
    parts: List[str] = get_toplevel(header)
    to_expand = list(
        filter(lambda x: x[1].startswith("for_each_sample"), enumerate(parts))
    )
    if len(to_expand) > 0 and vcf is None:
        raise ValueError("If FORMAT is to be expanded, the VCF kwarg must not be none.")
    parts_expanded: List[List[str]] = [[p] for p in parts]
    func = (
        generate_for_each_sample_expressions
        if make_expression
        else generate_for_each_sample_column_names
    )
    for i, p in to_expand:
        expanded = func(p, vcf)
        parts_expanded[i] = expanded
    parts_flattened = [p for pp in parts_expanded for p in pp]
    return ", ".join(parts_flattened)


def get_header(args, vcf: VariantFile) -> List[str]:
    if args.header == "auto":
        header = args.expression
    else:
        header = args.header
    if args.long:
        header = f"SAMPLE, {header}"
    return get_toplevel(preprocess_expression(header, vcf, args.header == "auto"))


def get_toplevel(header: str) -> List[str]:
    splitpos = [0]
    level = 0
    stack = []
    for i, c in enumerate(header):
        if c == "," and level == 0:
            splitpos.append(i + 1)
        elif c == "(":
            stack.append("(")
            level += 1
        elif c == ")":
            previous = stack[-1]
            if previous == "(":
                # found matching bracket pair
                stack.pop()
                level -= 1
            else:
                raise SyntaxError("No matching ( found.")
    if len(stack) != 0:
        raise SyntaxError("Imbalanced number of brackets.")
    splitpos.append(len(header) + 1)
    parts = []
    for start, end in zip(splitpos, splitpos[1:]):
        # remove leading + trailing whitespace
        parts.append(header[start : end - 1].strip())
    return parts


def get_row(row):
    if not isinstance(row, tuple):
        row = (row,)
    return row


@contextlib.contextmanager
def smart_open(filename=None, *args, **kwargs):
    if filename and filename != "-":
        fh = open(filename, *args, **kwargs)
    else:
        fh = sys.stdout

    try:
        yield fh
    finally:
        if fh is not sys.stdout:
            fh.close()


def execute(args):
    aux = read_auxiliary(args.aux)
    with VariantFile(args.vcf) as vcf:
        expression = preprocess_expression(args.expression, vcf, True)
        if args.long:
            expression = f"SAMPLE, {expression}"
        overwrite_number = {
            "INFO": dict(args.overwrite_number_info),
            "FORMAT": dict(args.overwrite_number_format),
        }
        rows = tableize_vcf(
            vcf,
            expression,
            args.annotation_key,
            overwrite_number=overwrite_number,
            long=args.long,
            auxiliary=aux,
        )

        try:
            with smart_open(args.output, "wt", newline="") as csvfile:
                writer = csv.writer(
                    csvfile, delimiter=args.separator, quoting=csv.QUOTE_MINIMAL
                )
                if args.header != "none":
                    header = get_header(args, vcf)
                    n_header_cols = len(header)
                    expr_cols = get_toplevel(expression)
                    n_expr_cols = len(expr_cols)
                    if n_header_cols != n_expr_cols:
                        raise HeaderWrongColumnNumber(
                            n_expr_cols, expr_cols, n_header_cols, header
                        )
                    writer.writerow(header)
                writer.writerows(get_row(row) for row in rows)
        except VembraneError as ve:
            print(ve, file=stderr)
            exit(1)



import math
import re
import statistics
from typing import Any, Callable, Dict, Iterable, Iterator, Optional, TypeVar, Union

from .ann_types import NA, NoValue

# The builtins list below was generated with:
#    python -c 'print(
#        *sorted(o for o in dir(__builtins__) if o.islower() and not o.startswith("_")),
#        sep="\n",
#    )'
_builtins = {
    obj.__name__: obj
    for obj in (
        abs,
        all,
        any,
        # # ascii,
        # bin,
        bool,
        # # breakpoint,
        # # bytearray,
        # # bytes,
        # # callable,
        chr,
        # # classmethod,
        # # compile,
        # # complex,
        # # copyright,
        # # credits,
        # # delattr,
        dict,
        # # dir,
        # divmod,
        enumerate,
        # # eval,
        # # exec,
        # # exit,
        filter,
        float,
        # format,
        # frozenset,
        # # getattr,
        # # globals,
        # # hasattr,
        # # hash,
        # # help,
        # hex,
        # # id,
        # # input,
        int,
        # isinstance,
        # # issubclass,
        iter,
        len,
        # # license,
        list,
        # # locals,
        map,
        max,
        # # memoryview,
        min,
        next,
        # # object,
        # oct,
        # # open,
        ord,
        # pow,
        # # print,
        # # property,
        # # quit,
        range,
        # # repr,
        reversed,
        round,
        set,
        # # setattr,
        # # slice,
        sorted,
        # # staticmethod,
        str,
        sum,
        # # super,
        tuple,
        # # type,
        # # vars,
        zip,
    )
}


_modules = {mod.__name__: mod for mod in (re,)}

_math_exports = {
    name: mod for name, mod in vars(math).items() if not name.startswith("__")
}


_statistics_exports = {
    name: mod
    for name, mod in vars(statistics).items()
    if name in statistics.__all__ and name[0].islower()
}


T = TypeVar("T")


def without_na(values: Iterable[Union[T, NoValue]]) -> Iterator[T]:
    """Keep only values that are not `NA`."""
    return filter(lambda v: v is not NA, values)


def replace_na(values: Iterable[Union[T, NoValue]], replacement: T) -> Iterator[T]:
    """Replace values that are `NA` with `replacement`."""
    for v in values:
        if v is not NA:
            yield v
        else:
            yield replacement


_additional_functions = {
    "without_na": without_na,
    "replace_na": replace_na,
}

_explicit_clear: Dict[str, Optional[Dict[str, Any]]] = {
    "__builtins__": {},
    "__builtin__": {},
    "__file__": None,
    "__name__": None,
    "__doc__": None,
    "__package__": None,
    "__import__": None,
}

allowed_globals: Dict[str, Any] = {
    **_builtins,
    **_modules,
    **_math_exports,
    **_statistics_exports,
    **_additional_functions,
    **_explicit_clear,
    "NA": NA,
}


def custom_functions(env) -> Dict[str, Callable]:
    return {
        "count_hom": eval(
            "lambda: "
            "sum(all(x == FORMAT['GT'][s][0] for x in FORMAT['GT'][s][1:]) "
            "for s in SAMPLES)",
            env,
            {},
        ),
        "count_het": eval(
            "lambda: "
            "sum("
            "any(x != next(f for f in FORMAT['GT'][s] if f is not NA) "
            "for x in FORMAT['GT'][s][1:] if x is not NA) "
            "for s in SAMPLES)",
            env,
            {},
        ),
        "count_any_ref": eval(
            "lambda: sum(any(x == 0 for x in FORMAT['GT'][s]) for s in SAMPLES)",
            env,
            {},
        ),
        "count_any_var": eval(
            "lambda: "
            "sum(any(x != 0 for x in FORMAT['GT'][s] if x is not NA) "
            "for s in SAMPLES)",
            env,
            {},
        ),
        "count_hom_ref": eval(
            "lambda: sum(all(x == 0 for x in FORMAT['GT'][s]) for s in SAMPLES)",
            env,
            {},
        ),
        "count_hom_var": eval(
            "lambda: sum(all(x != 0 and x is not NA for x in FORMAT['GT'][s]) "
            "for s in SAMPLES)",
            env,
            {},
        ),
        "is_hom": eval(
            "lambda sample: "
            "all(x == next(f for f in FORMAT['GT'][sample] if f is not NA) "
            "for x in FORMAT['GT'][sample][1:])",
            env,
            {},
        ),
        "is_het": eval(
            "lambda sample: "
            "any(x != next(f for f in FORMAT['GT'][sample] if f is not NA) "
            "for x in FORMAT['GT'][sample][1:] if x is not NA)",
            env,
            {},
        ),
        "is_hom_ref": eval(
            "lambda sample: all(x == 0 for x in FORMAT['GT'][sample])",
            env,
            {},
        ),
        "is_hom_var": eval(
            "lambda sample: all(x != 0 and x is not NA for x in FORMAT['GT'][s])",
            env,
            {},
        ),
        "has_ref": eval(
            "lambda sample: any(x == 0 for x in FORMAT['GT'][sample])",
            env,
            {},
        ),
        "has_var": eval(
            "lambda sample: "
            "any(x != 0 for x in FORMAT['GT'][sample] if x is not NA)",
            env,
            {},
        ),
    }
